/*
 * Copyright (C) 2002-2005 StarBED Project All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 *  KOTETU: hostname resolution procedures.
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995
 *
 *  $Id: hname.c,v 1.1 2006/11/27 01:56:36 k-chinen Exp $
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#if 0
#include "range.h"
#include "msg.h"
#endif

#include "msg.h"
#if 0
#include "kcomp.h"
#endif

#define Trace   RawTrace

int
raw_show_addr(char *prefix, unsigned char *ad)
{
    Trace("%s: IPv4addr decimal %u.%u.%u.%u. (hex %02x%02x%02x%02x)\n", 
        prefix, ad[0], ad[1], ad[2], ad[3], ad[0], ad[1], ad[2], ad[3]);
}

int
old_raw_show_addr(char *prefix, struct in_addr *addr)
{
    unsigned long mid;
    unsigned char *ad;

    mid = ntohl(*(unsigned long*)addr);
    ad = (unsigned char*)mid;

    Trace("%s: IPv4addr decimal %u.%u.%u.%u. (hex %02x%02x%02x%02x)\n", 
        prefix, ad[0], ad[1], ad[2], ad[3], ad[0], ad[1], ad[2], ad[3]);

    mid = *(unsigned long*)addr;
    ad = (unsigned char*)mid;

    Trace("%s: IPv4addr decimal %u.%u.%u.%u. (hex %02x%02x%02x%02x)\n", 
        prefix, ad[0], ad[1], ad[2], ad[3], ad[0], ad[1], ad[2], ad[3]);
}

/*
 * addr_byname - find address structure by name
 *
 * return:
 *      1   error
 *      0   success
 */
int addr_byname(char *hostname, int portno, struct sockaddr_in *addr)
{
    struct hostent  *hp;

#if 0
    Trace("addr_byname: host-'%s'\n", hostname);
#endif

    if(hostname[0]=='\0') {
        return 1;
    }
    if((hp=gethostbyname(hostname)) == NULL ) {
        Trace("addr_byname: not found host-'%s'\n", hostname);
        return 1;
    }

    memset(addr, 0, sizeof(addr));
    addr->sin_family = hp->h_addrtype;
    addr->sin_port = htons(portno);
    memcpy(&(addr->sin_addr.s_addr), hp->h_addr, hp->h_length);

#if 0
    Trace("addr_byname: found host-'%s'\n", hostname);
#endif

    return 0;
}


/*
 * addr_bynumber - find address structure by digit number for IP address
 *
 * return:
 *      1   error
 *      0   success
 */
int addr_bynumber(char *hostname, int portno, struct sockaddr_in *addr)
{
#if 1
    char xhost[BUFSIZ];
    char *p, *q;
#endif
    /* support only digit */
    if(hostname[0]=='\0' || !isdigit(hostname[0])) {
        Trace("addr_bynumber: ignore digit address-'%s'\n", hostname);
        return 1;
    }

    memset(addr, 0, sizeof(addr));
    addr->sin_family = AF_INET;
    addr->sin_port = htons(portno);
#if 0
    addr->sin_addr.s_addr = inet_addr(hostname);
#endif
#if 1
    /* remove chars after '/' */
    p = hostname;
    q = xhost;
    while(*p) {
        if(*p=='/') {
            break;
        }
        *q++ = *p++;
    }
    *q = '\0';
#if 0
    fprintf(stderr, "hostname '%s' xhost '%s'\n", hostname, xhost);
#endif
    addr->sin_addr.s_addr = inet_addr(xhost);
    if(addr->sin_addr.s_addr==-1) {
#if 0
        fprintf(stderr, "fail inet_addr()\n");
#endif
        return -1;
    }
    else {
#if 0
        fprintf(stderr, "success inet_addr()\n");
#endif
    }
#endif

    return 0;
}


/*
 * find_addr - find address structure by string
 *
 * return:
 *      1   error
 *      0   success
 */
int find_addr(char *hostname, int portno, struct sockaddr_in *addr)
{
    /* support only digit */
    if(hostname[0]=='\0') {
        return 1;
    }

#if 0
    if(fetch_hostname_cache(hostnamem, addr)) {
        return 0;
    }
#endif

    if(isdigit(hostname[0])) {
        return addr_bynumber(hostname, portno, addr);
    }
    else {
        return addr_byname(hostname, portno, addr);
    }
}


