/*
 * Copyright (C) 2002-2005 StarBED Project All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 *  KOTETU: signals procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1995, 1996
 *
 *  $Id: sigs.c,v 1.1 2006/11/27 01:56:36 k-chinen Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/uio.h>
#include <sys/socket.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif

#include <errno.h>

#ifndef KOTETU_EXIT_NORMAL
#define KOTETU_EXIT_NORMAL  (0)
#endif

#if 0
#include "range.h"
#include "proxy_global.h"
#endif
#include "msg.h"

#define Trace   RawTrace
#define DLog    Error
#define Log     Error


#include "sigs.h"


volatile sig_atomic_t pipe_broken;
volatile sig_atomic_t catch_termination;


void SetSignalHandler(int signo, RETSIGTYPE (*handler)(int))
{
#ifdef HAVE_SIGACTION
    struct sigaction act, oact;

    act.sa_handler = handler;
    sigemptyset(&act.sa_mask);

    act.sa_flags = 0;

#if 0

#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;
#else
#ifdef SV_INTERRUPT
    act.sa_flags |= SV_INTERRUPT;
#endif
#endif

#else

    if(signo == SIGALRM) {
#ifdef SV_INTERRUPT
        act.sa_flags |= SV_INTERRUPT;
#endif
    }
    else {
#ifdef SA_RESTART
        act.sa_flags |= SA_RESTART;
#endif
    }

#endif

    sigaction(signo, &act, &oact);

#else
    /* no sigaction() */
    signal(signo, handler);

#endif /* HAVE_SIGACTION */
}




/*
 *
 */
RETSIGTYPE PipeErrorHandler()
{
    Trace("*** PIPE pipe was broken\n");

    pipe_broken=1;

    return;
}



/*
 *
 */
RETSIGTYPE ReleseChild()
{
    pid_t pid;

#ifdef HAVE_WAIT3

#ifdef HAVE_SYS_WAIT_H      /* like POSIX */
    int status;
#else
    union wait status;
#endif

    while((pid = wait3(&status, WNOHANG | WUNTRACED,
        (struct rusage*)NULL)) > 0 ) {

        DLog("INFO chlid %d (unknown*) die.", pid);
    }

#else /* HAVE_WAIT3 */

#ifdef HAVE_SYS_WAIT_H      /* like POSIX */
    while((pid = wait((int*) NULL)) >= 0 ) {
        DLog("INFO chlid %d (unknown*) die.", pid);
    }
#else
    while((pid = wait((union wait*) NULL)) >= 0 ) {
        DLog("INFO chlid %d (unknown*) die.", pid);
    }
#endif

#endif /* HAVE_WAIT3 */
}



/*
 *
 */
RETSIGTYPE ExitProgram()
{
    Log("; Catch signal for exit. goodbye !");
    exit(KOTETU_EXIT_NORMAL);
}



/*
 *
 */
RETSIGTYPE HookTERM()
{
    Log("; Catch signal for termination");
    catch_termination = 1;
}



/*
 *
 */
RETSIGTYPE DoNothing()
{
}



