/*
 * Copyright (C) 2002-2005 StarBED Project All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * string manipulate
 *
 * $Id: sman.c,v 1.1 2006/11/27 01:56:36 k-chinen Exp $
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>

#include "msg.h"

volatile int trace_sman=0;

#ifndef Error
#define Error   printf
#endif
#ifndef Trace
#define Trace   if(trace_sman)RawTrace
#endif
#ifndef Echo
#define Echo    RawTrace
#endif

#ifndef PASS
#define PASS    \
    fprintf(stderr, "PASS %s:%d\n",__FILE__,__LINE__); fflush(stderr)
#endif

#ifndef VLSTR_LEN
#define VLSTR_LEN    (BUFSIZ*4)
#endif
#ifndef MSTR_LEN
#define MSTR_LEN    (BUFSIZ)
#endif
#ifndef SSTR_LEN
#define SSTR_LEN    (64)
#endif



char*
Chop_CRLF(char *line)
{
    char *p;

    if(line==NULL) {
        return NULL;
    }

    p = line;
    while(*p) {
        if(*p=='\r' && *(p+1)=='\n' && *(p+2)=='\0') {
            *p = '\0';
            break;
        }
        if(*p=='\n' && *(p+1)=='\0') {
            *p = '\0';
            break;
        }
        p++;
    }

    return line;
}

int
Capitalize(char *line)
{
    register char *p;

    p = line;
    while(*p) {
        *p = toupper(*p);
        p++;
    }

    return 0;
}

/*
 *
 */
char *sgets(char *pos, char *buf, int blen)
{
    register char *p, *q;
    register int  c;

    if(pos==NULL) {
        return NULL;
    }

#if 0
    dump_string("      rest", pos);
#endif

    q = buf;
    p = pos;
    c = 0;

    while(*p&&c<blen) {
        c++;
        if(*p=='\n') {
            *q++ = *p++;
            *q = '\0';
            return p;
        }
        *q++ = *p++;
    }
    *q = '\0';

    if(p==pos)
        return NULL;
    else
        return p;
}

/*
 * dump string segment
 */
void
dump_stringsegment(char *prefix, char *str, int len)
{
    char lline[VLSTR_LEN];
    register int c, i;
    register char *p, *q;

    if(str==NULL || len<=0) {
        return;
    }

#if 0
    /* range check */
    p = str;
    q = lline;
    for(i=0;i<len;i++) {
        *q = *p;
        p++;
    }
#endif

    lline[0] = '\0';
    strcat(lline, prefix);
    q = lline;
    c = 0;
    while(*q) {
        if(c>=VLSTR_LEN-1) {
            *q = '\0';
            Echo("%s", lline);
            q = lline;
            c = 0;
        }
        q++;
        c++;
    }

    *q++ = '[';
    c++;

    p = str;
    i = 0;
    while(i<len) {
        if(c>=VLSTR_LEN-1) {
            *q = '\0';
            Echo("%s", lline);
            q = lline;
            c = 0;

            continue;
        }

        if(*p=='\0')  {
            *q++ = '{';
            *q++ = 'N';
            *q++ = 'U';
            *q++ = 'L';
            *q++ = 'L';
            *q++ = '}';
        }
        else
        if(*p=='\r')  {
            *q++ = '{';
            *q++ = 'C';
            *q++ = 'R';
            *q++ = '}';
        }
        else
        if(*p=='\n')  {
            *q++ = '{';
            *q++ = 'L';
            *q++ = 'F';
            *q++ = '}';
        }
        else {
            if(*p<' ' || *p>0x7f) {
                *q++ = '{';
                *q++ = "01234567890abcdefz"[*p/16];
                *q++ = "01234567890abcdefz"[*p%16];
                *q++ = '}';
            }
            else {
                *q++ = *p;
            }
        }

        p++;
        i++;
        c++;
    }

    *q++ = ']';
    c++;

    *q++ = '\n';
    c++;

    if(c>0) {
        *q = '\0';
        Echo("%s", lline);
        q = lline;
        c = 0;
    }

}


/*
 * dump string line
 */

void
dump_string(char *prefix, char *line)
{
    unsigned char *p;

    Echo("%s[", prefix);
    p = (unsigned char*) line;
    if(*p) {
        while(*p&&*p!='\r'&&*p!='\n') {
            if(*p<' ' || *p>0x7f) {
                Echo("{%02x}",*p);
            }
            else {
                Echo("%c",*p);
            }
            p++;
        }
        if(*p=='\0')
            Echo("{NULL}");

        if(*p=='\r') {
            Echo("{CR}");
            p++;
        }
        if(*p=='\n')
            Echo("{LF}");
        Echo("]\n");
    }
    else {
        Echo("{EMPTY}]\n");
    }
}

/*
 * dump string chunk
 */
void
dump_stringchunk(char *prefix, char *chunk)
{

#define concat          \
            {\
                t = (unsigned char*)tmp;\
                while(m<MSTR_LEN && *t) {\
                    *q++ = *t++;\
                    m++;\
                }\
                *q = '\0';\
            }

    register unsigned char *p, *q, *t;
    register int m;
    int  c; /* line number */
    char tmp[16];
    char label[20];

#if 0
    unsigned char buf[MSTR_LEN];
#endif
#if 1
    unsigned char *buf;
    buf = (unsigned char*)malloc(MSTR_LEN);
    if(buf==NULL) {
        Trace("nomemory\n");
    }
#endif

    c = 0;
    p = (unsigned char*) chunk;
    if(!*p) {
        Echo("%s %03d[{EMPTY}]\n", prefix, c);
        return;
    }

    if(strlen(prefix)>=20-1) {
        strncpy(label, prefix, 20);
        label[20] = '\0';
    }
    else {
        strcpy(label, prefix);
    }

    while(*p) {
        c++;

        memset(buf, 0, MSTR_LEN);
        q = buf;
        m = 0;

        while(*p&&*p!='\r'&&*p!='\n') {
            if(*p<' ' || *p>=0x7f) {
                sprintf(tmp, "{%02x}",*p);
                concat;
            }
            else {
                *q++ = *p;
                m++;
                *q = '\0';
            }
            p++;
        }
        if(*p=='\0') {
            concat;
        }

        if(*p=='\r') {
            sprintf(tmp, "{CR}");
            concat;
            p++;
        }
        if(*p=='\n') {
            sprintf(tmp, "{LF}");
            concat;
            p++;
        }

        Echo("%s %02d %03d [%s]\n", label, c, m, buf);
        if(c==1) {
            /* fill space in label */
            t = (unsigned char*)label; while(*t) { *t = ' '; t++; }
        }
    }
#if 1
    if(buf)
        free(buf);
#endif

#undef concat
}




/*****
 *****
 *****/

static
int
split_hpp(char *src, char *prefix, int *head, int *tail)
{
    char *p, *q;
    char tmp[BUFSIZ];
    int c;

    *head = -1;
    *tail = -1;

    p = src;
    q = prefix;
    c = 0;
    while(c<BUFSIZ&&*p&&((*p>='a'&&*p<='z')||(*p>='A'&&*p<='Z'))) {
        *q++ = *p++;
        c++;
    }
    *q = '\0';

    q = tmp;
    while(*p&&*p>='0'&&*p<='9') {
        *q++ = *p++;
    }
    *q = '\0';
    *head = atoi(tmp);

    if(*p!='-') {
        *tail = *head;
        goto out;
    }

    p++;

    q = tmp;
    while(*p&&*p>='0'&&*p<='9') {
        *q++ = *p++;
    }
    *q = '\0';
    *tail = atoi(tmp);

out:
#if 0
    Trace("split_hpp: src '%s' -> '%s' %d %d\n",
        src, prefix, *head, *tail);
#endif

    return 0;
}


int
expand_hpat(char *dst, int dmax, char *pat)
{
    char *p, *q;
    char h[BUFSIZ];
    int ck;
    int hc;
    char nh[BUFSIZ];
    char pre[BUFSIZ];
    int st, en;
    int i;
    int dlen;
    int hlen;
    int ret;
    

    Trace("expand_hpat: '%s'\n", pat);

    ret = 0;
    dlen = 0;
    dst[0] = '\0';

    p = pat;
    hc = 0;
    while(*p) {
        q = h;
        hlen = 0;
        while(hlen<BUFSIZ && *p && *p!=',') {
            *q++ = *p++;
            hlen++;
        }
        *q = '\0';

#if 1
        Trace(" dlen %d/%d\n", dlen, dmax);
        Trace(" hpart %3d '%s' <%d>\n", hc, h, hlen);
#endif
        if(strcmp(h, "localhost")==0 && dlen+hlen<dmax) {
            strcpy(dst, h);
            dlen += hlen;
            
            hc++;
        }
        else {
            ck = split_hpp(h, pre, &st, &en);
#if 1
            Trace("  pre '%s', st %d, en %d\n", pre, st, en);
#endif
            for(i=st;i<=en;i++) {
                if(! (strlen(pre)+4<BUFSIZ)) {
                    fprintf(stderr, "too long basename part <%d>\n",
                        strlen(pre));
                    fflush(stderr);
                    continue;
                }

                if( (pre[0]>='a'&&pre[0]<='f') && (pre[1]=='\0')) {
#if 0
                    Trace("   shortname\n");
#endif
                    sprintf(nh, "sintcl%s%03d", pre, i);
                }
                else {
#if 0
                    Trace("   fullname\n");
#endif
                    sprintf(nh, "%s%03d", pre, i);
                }
                hlen = strlen(nh);
#if 0
                Trace("  '%s' <%d>\n", nh, hlen);
#endif
                if(hc!=0) {
                    if(dlen+1<dmax) {
                        strcat(dst, ",");
                        dlen++;
                    }
                }
#if 0
                Trace(" len.check %d+%d<%d ? %d\n",
                    dlen, hlen, dmax, (dlen+hlen<dmax));
#endif
                if(dlen+hlen<dmax) {
                    strcat(dst, nh);
                    dlen += hlen;
                }
                else {
                    ret = 1;
                    goto out;
                }

                hc++;
            }
        }

        if(*p==',') {
            p++;
        }
    }

out:
#if 0
    Trace("dst <%d> '%s'\n", dlen, dst);
    fflush(stdout);
#endif

    return ret;
}


static
char*
skipwhite(char *p)
{
    while(*p==' ' ||*p=='\t') {
        p++;
    }
    return p;
}

int
parse_cmd(char *line, int aplen, int amax, char *ap)
{
    register int i, j;
    register char *p, *q;
    int c;

#if 0
    if(dlevel>3) {
        dump_string("parsecmd", line);
    }
#endif
    
    p = line;
    i = 0;
    while(i<amax) {
        p = skipwhite(p);
        if(*p=='\0') {
            break;
        }
        j = 0;
        q = &ap[aplen*i+j];
        while(j<aplen && *p && *p!=' ' && *p!='\t'
                && *p!='\r' && *p!='\n') {
            *q++ = *p++;
            j++;
        }
        *q = '\0';
        i++;
        if(*p=='\0') {
            break;
        }
    }

    c = i;
    
#if 0
    for(i=0;i<c;i++) {
        Trace("'%s'\n", &ap[aplen*i+0]);
    }
#endif
    
    return c;
}



