/*
 * Copyright (C) 2002-2005 StarBED Project All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 *  KOTETU: socket (TCP/UDP) procedures
 *      by k-chinen@is.aist-nara.ac.jp, 1994, 1995, 1996
 *
 * $Id: socket.h,v 1.1 2006/11/27 01:56:36 k-chinen Exp $
 */

#ifndef KOTETU_SOCKET_H
#define KOTETU_SOCKET_H

#ifndef KOTETU_LISTEN_LENGTH
#define KOTETU_LISTEN_LENGTH      (128)
#endif /* KOTETU_LISTEN_LENGTH */


#define KOTETU_SOCK_SOCKETERROR   (-1)
#define KOTETU_SOCK_UNKNOWHOST    (-2)
#define KOTETU_SOCK_CONNECTERROR  (-3)


int bind_listen_port(int type, int portno, int listen_queue_length);

#ifndef DONTUSE_POLL
#ifdef  HAVE_POLL
#ifndef USE_POLL
#define USE_POLL
#endif /* USE_POLL */
#else
#ifdef USE_POLL
#undef USE_POLL
#endif /* USE_POLL */
#endif /* HAVE_POLL */
#endif /* DONUSE_POLL */

int lock_port(int pfd);
int unlock_port(int pfd);

int isreadable_via_poll(int fd, int sec, int usec);
int iswritable_via_poll(int fd, int sec, int usec);
int isreadable_via_select(int fd, int sec, int usec);
int iswritable_via_select(int fd, int sec, int usec);
#ifdef USE_POLL
#define isreadable(f, s, u)     isreadable_via_poll(f,s,u)
#define iswritable(f, s, u)     iswritable_via_poll(f,s,u)
#else
#define isreadable(f, s, u)     isreadable_via_select(f,s,u)
#define iswritable(f, s, u)     iswritable_via_select(f,s,u)
#endif

#ifndef US
#define US(x)       (1L*(x))
#define MS(x)       (1000L*(x))
#define S(x)        (1000L*1000L*(x))
#endif
int waitread_via_poll(int fd,   int usec);
int waitread_via_select(int fd, int usec);
#ifdef USE_POLL
#define waitread(f, u)     waitread_via_poll(f,u)
#else
#define waitread(f, u)     waitread_via_select(f,u)
#endif


int sync_connect_byname(int type, char *hostname, int portno);
int sync_connect_byaddr(int type, struct sockaddr_in *addr, int portno);
int async_connect_byname(int type, char *hostname, int portno);
int async_connect_byaddr(int type, struct sockaddr_in *addr, int portno);

int disable_nb(int);
int release_socket(int *sfd);


/*
 * synatx sugers
 */

/* UDP */
#define UDP_bind_port(portno, length) \
            bind_listen_port(SOCK_DGRAM, portno, length)
#define UDP_connect(hostname, portno) \
            sync_connect_byname(SOCK_DGRAM, hostname, portno)
#define UDP_aconnect(hostname, portno) \
            async_connect_byname(SOCK_DGRAM, hostname, portno)
#define UDP_connect_addr(addr, portno) \
            sync_connect_byaddr(SOCK_DGRAM, addr, portno)
#define UDP_aconnect_addr(addr, portno) \
            async_connect_byaddr(SOCK_DGRAM, addr, portno)

/* TCP */
#define TCP_bind_port(portno, length) \
            bind_listen_port(SOCK_STREAM, portno, length)
#define TCP_connect(hostname, portno) \
            sync_connect_byname(SOCK_STREAM, hostname, portno)
#define TCP_aconnect(hostname, portno) \
            async_connect_byname(SOCK_STREAM, hostname, portno)
#define TCP_connect_addr(addr, portno) \
            sync_connect_byaddr(SOCK_STREAM, addr, portno)
#define TCP_aconnect_addr(addr, portno) \
            async_connect_byaddr(SOCK_STREAM, addr, portno)


#endif /* KOTETU_SOCKET_H */
